/*
Copyright (C) 2014 Jacquelin POTIER <jacquelin.potier@free.fr>
Dynamic aspect ratio code Copyright (C) 2014 Jacquelin POTIER <jacquelin.potier@free.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//-----------------------------------------------------------------------------
// Object: class helper for hot window (mouse trap for auto show / auto hide window)
//-----------------------------------------------------------------------------

#include "HotWindow.h"
#include "../DesktopWorkArea/DesktopWorkArea.h"

#define CHotWindow_MAX_LONG 0x7fffffff
#define CHotWindow_DEFAULT_AUTO_SHOW_TIME 50
#define CHotWindow_DEFAULT_AUTO_HIDE_TIME 500

CHotWindow::CHotWindow(IHotWindowEvents* pEventsHandler)
{
    this->bAutoShowStarted = FALSE;
    this->bAutoHideStarted = FALSE;
    this->hWnd = NULL;
    this->RegisteredhWnd = NULL;
    this->pEventsHandler = pEventsHandler;
    this->AutoShowTime = CHotWindow_DEFAULT_AUTO_SHOW_TIME;
    this->AutoHideTime = CHotWindow_DEFAULT_AUTO_HIDE_TIME;
    this->bUseAutoHideDisabledExtraRect = FALSE;
    memset(&this->AutoHideDisabledExtraRect,0,sizeof(RECT));

    // Initialize rect and old point with unreachable value
    // Notice : in case of multiple screen cursor position and window rect can have negative values
    this->ActiveRect.left = CHotWindow_MAX_LONG;
    this->ActiveRect.right = CHotWindow_MAX_LONG;
    this->ActiveRect.top = CHotWindow_MAX_LONG;
    this->ActiveRect.bottom = CHotWindow_MAX_LONG;

    this->OldCursorPos.x = CHotWindow_MAX_LONG;
    this->OldCursorPos.y = CHotWindow_MAX_LONG;
}
CHotWindow::~CHotWindow()
{
    this->StopAutoShow();
    this->StopAutoHide();
}

BOOL CALLBACK CHotWindow::EnumWindowsProc(HWND hWndCurrentWindow,LPARAM lParam)
{
    tagEnumWindowParam* pParam = (tagEnumWindowParam*)lParam;
    if (!::IsWindowVisible(hWndCurrentWindow))
        return TRUE;
    // Notice : ForeGroundWindow is not enough so we need to pass through enum window
    //          as some applications (like m$ media player in full screen mode) don't return the full screen window handle
    // HWND hWndCurrentWindow = ::GetForegroundWindow();
    if( (hWndCurrentWindow != ::GetDesktopWindow()) && (hWndCurrentWindow != ::GetShellWindow()) )
    {
        RECT CurrentAppRect={0};
        RECT ScreenRect={0};
        DWORD ZeroBasedMonitorIndex;
        ::GetWindowRect(hWndCurrentWindow, &CurrentAppRect);
        if (::PtInRect(&CurrentAppRect,pParam->CursorPos))
        {
            if (CDesktopWorkArea::GetWorkAreaForMonitorFromPoint(&pParam->CursorPos,&ZeroBasedMonitorIndex,&ScreenRect,TRUE))
            {
                // if screen is fully included in application rect
                // note : do not use EqualRect, as some application rect can be larger than screen
                // bDoShow = !::EqualRect(&ScreenRect,&CurrentAppRect);
                BOOL bScreenFullyIncluded =    (CurrentAppRect.left<=ScreenRect.left) && (CurrentAppRect.right>=ScreenRect.right)
                                            && (CurrentAppRect.top<=ScreenRect.top) && (CurrentAppRect.bottom>=ScreenRect.bottom);
                pParam->bAWindowIsInFullScreenMode = bScreenFullyIncluded;
            }
        }
    }

    // as soon as we found an full screen window, we can stop parsing
    if (pParam->bAWindowIsInFullScreenMode)
        return FALSE;
    else
        return TRUE;
}

BOOL CHotWindow::DoesScreenContainingMouseHasAFullScreenWindow(POINT* pCursorPos)
{
    tagEnumWindowParam Param;
    Param.CursorPos = *pCursorPos;
    Param.bAWindowIsInFullScreenMode = FALSE;
    ::EnumWindows(CHotWindow::EnumWindowsProc,(LPARAM)&Param);
    return Param.bAWindowIsInFullScreenMode;
}

void CALLBACK CHotWindow::TimerProcAutoShow(HWND hWnd,UINT uMsg,UINT_PTR idEvent,DWORD dwTime)
{
    UNREFERENCED_PARAMETER(hWnd);
    UNREFERENCED_PARAMETER(uMsg);
    UNREFERENCED_PARAMETER(dwTime);
    CHotWindow* pHotWindow = (CHotWindow*)idEvent;

    // The KillTimer function does not remove WM_TIMER messages already posted to the message queue
    if (!pHotWindow->bAutoShowStarted)
        return;

    POINT CursorPos;
    ::GetCursorPos(&CursorPos);
    if (::PtInRect(&pHotWindow->ActiveRect,CursorPos))
    {
        BOOL bDoShow = FALSE;
        // assume mouse doesn't move too much (no more than a double click) between last position and the current one
        LONG dx = CursorPos.x-pHotWindow->OldCursorPos.x;
        LONG dy = CursorPos.y-pHotWindow->OldCursorPos.y;
        LONG MovedDistance = dx*dx + dy*dy;
        LONG MaxDoubleClickDistance = ::GetSystemMetrics(SM_CXDOUBLECLK) * ::GetSystemMetrics(SM_CYDOUBLECLK);
        if (MovedDistance <= MaxDoubleClickDistance)
            bDoShow = TRUE;

        // win 10 and win 11 have full screen window process like TestInputHost.exe, so don't use it
        //if (bDoShow)
        //{
        //    // assume an application is not running in full screen mode
        //    bDoShow = !CHotWindow::DoesScreenContainingMouseHasAFullScreenWindow(&CursorPos);
        //}

        if (bDoShow)
        {
            pHotWindow->StopAutoShow();
            pHotWindow->StartAutoHide();
            if (pHotWindow->pEventsHandler)
                pHotWindow->pEventsHandler->OnHotWindowAutoShow(pHotWindow,&pHotWindow->bUseAutoHideDisabledExtraRect,&pHotWindow->AutoHideDisabledExtraRect);
        }
    }
    pHotWindow->OldCursorPos = CursorPos;
}

void CALLBACK CHotWindow::TimerProcAutoHide(HWND hWnd,UINT uMsg,UINT_PTR idEvent,DWORD dwTime)
{
    UNREFERENCED_PARAMETER(hWnd);
    UNREFERENCED_PARAMETER(uMsg);
    UNREFERENCED_PARAMETER(dwTime);
    CHotWindow* pHotWindow = (CHotWindow*)idEvent;

    // The KillTimer function does not remove WM_TIMER messages already posted to the message queue
    if (!pHotWindow->bAutoHideStarted)
        return;

    POINT CursorPos;
    ::GetCursorPos(&CursorPos);
    BOOL bCursorInsideActiveRegions = FALSE;

    // check extra rect if required
    if (pHotWindow->bUseAutoHideDisabledExtraRect)
    {
        if ( (::PtInRect(&pHotWindow->AutoHideDisabledExtraRect,CursorPos)) || (::PtInRect(&pHotWindow->AutoHideDisabledExtraRect,pHotWindow->OldCursorPos)) )
            bCursorInsideActiveRegions = TRUE;
    }

    // check active rect
    if (!bCursorInsideActiveRegions)
    {
        if ( (::PtInRect(&pHotWindow->ActiveRect,CursorPos)) || (::PtInRect(&pHotWindow->ActiveRect,pHotWindow->OldCursorPos)) )
            bCursorInsideActiveRegions = TRUE;
    }

    if (!bCursorInsideActiveRegions)
    {
        pHotWindow->StopAutoHide();
        pHotWindow->StartAutoShow();
        if (pHotWindow->pEventsHandler)
            pHotWindow->pEventsHandler->OnHotWindowAutoHide(pHotWindow);
    }

    pHotWindow->OldCursorPos = CursorPos;
}

BOOL CHotWindow::SetWindowAndActiveRect(HWND hWnd,RECT* pActiveRect)
{
    if (!::IsWindow(hWnd))
        return FALSE;

    this->hWnd = hWnd;
    this->ActiveRect = *pActiveRect;
    return TRUE;
}

void CHotWindow::SetAutoShowTime(UINT Time)
{
    this->AutoShowTime = Time;
}
void CHotWindow::SetAutoHideTime(UINT Time)
{
    this->AutoHideTime = Time;
}

BOOL CHotWindow::StartAutoShow()
{
    this->StopAutoHide();
    this->StopAutoShow();
    this->RegisteredhWnd = this->hWnd;
    // tricky stuff : use current object pointer has timer id to interact with current object in static timer proc
    if (!::SetTimer(this->RegisteredhWnd,(UINT_PTR)this,this->AutoShowTime,CHotWindow::TimerProcAutoShow))
        return FALSE;
    this->bAutoShowStarted = TRUE;
    return TRUE;
}
BOOL CHotWindow::StopAutoShow()
{
    if (!this->bAutoShowStarted)
        return TRUE;
    this->bAutoShowStarted = FALSE;
    return ::KillTimer(this->RegisteredhWnd,(UINT_PTR)this);
}
BOOL CHotWindow::StartAutoHide()
{
    this->StopAutoHide();
    this->StopAutoShow();
    this->RegisteredhWnd = this->hWnd;
    // tricky stuff : use current object pointer has timer id to interact with current object in static timer proc
    if (!::SetTimer(this->RegisteredhWnd,(UINT_PTR)this,this->AutoHideTime,CHotWindow::TimerProcAutoHide))
        return FALSE;
    this->bAutoHideStarted = TRUE;
    return TRUE;
}
BOOL CHotWindow::StopAutoHide()
{
    if (!this->bAutoHideStarted)
        return TRUE;
    this->bAutoHideStarted = FALSE;
    return ::KillTimer(this->RegisteredhWnd,(UINT_PTR)this);
}
